# EntDissolver by David Bowland
# ./addons/eventscripts/entdissolver/entdissolver.py

# >>> To configure this addon please see entdissolver.cfg (created when the addon is first loaded) <<<

"""
Dissolves idle weapons, ragdolls, and items.
Requires only ES 2.1+
"""

"""
   * EntDissolver makes the following functions available when imported:

   dissolve(target_index, dissolvetype, magnitude)
      - Dissolve an entity by index given a dissolve type and magnitude

   dissolveDelay(delay, target_index, dissolvetype, magnitude)
      - Dissolves an entity after a delay

   removeDelay(index)
      - Cancel the index delay and remove the index from our delay list

   cleanupAll()
      -Marks new idle ragdolls, weapons and items to be dissolved


   * To use EntDissolver functions in your addon, simply import it after it is loaded:

   entdissovler = es.import_addon('entdissolver')


   * You may then call the functions listed above:

   entdissolver.cleanupAll()
"""

import cfglib
import cmdlib
import es
import gamethread
import playerlib
import random
import weaponlib

from path import path

import psyco
psyco.full()

__all__ = ['dissolve', 'dissolveDelay', 'removeDelay', 'cleanupAll']


info = es.AddonInfo()
info.name     = 'EntDissolver'
info.version  = '1'
info.basename = 'entdissolver'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


addonpath = path(es.getAddonPath(info.basename))

###

config = cfglib.AddonCFG(addonpath.joinpath(info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')

config.text('***** Ragdoll options *****')
config.text('NOTE: The availability of dissolve types may be game-dependent')
cvar_ragdoll_delay     = config.cvar('entdissolver_ragdoll_delay',    6,      'Number of seconds ragdolls remain on the map, 0 = never dissolve ragdolls')
cvar_ragdoll_type      = config.cvar('entdissolver_ragdoll_type',     '0123', 'Dissolve types to use: 0 = Energy, 1 = Heavy electrical, 2 = Light electrical, 3 = Core effect')
cvar_ragdoll_magnitude = config.cvar('entdissolver_ragdoll_magnitude', 3,     'Magnitude of dissolve effect')
cvar_ragdoll_magnitude_score = config.cvar('entdissolver_ragdoll_magnitude_score', 1, '1 = magnitude of effect is multiplied by player score, 0 = magnitude of effect is unchanging')
config.text('\n')

config.text('***** Item options *****')
cvar_item_delay     = config.cvar('entdissolver_item_delay',     10,    'Number of seconds weapons remain on the map, 0 = never dissolve weapons')
cvar_item_type      = config.cvar('entdissolver_item_type',     '0123', 'Dissolve types to use: 0 = Energy, 1 = Heavy electrical, 2 = Light electrical, 3 = Core effect')
cvar_item_magnitude = config.cvar('entdissolver_item_magnitude', 3,     'Magnitude of dissolve effect')
config.text('\n')

config.text('***** Weapon options *****')
cvar_weapon_delay     = config.cvar('entdissolver_weapon_delay',     10,    'Number of seconds weapons remain on the map, 0 = never dissolve weapons')
cvar_weapon_type      = config.cvar('entdissolver_weapon_type',     '0123', 'Dissolve types to use: 0 = Energy, 1 = Heavy electrical, 2 = Light electrical, 3 = Core effect')
cvar_weapon_magnitude = config.cvar('entdissolver_weapon_magnitude', 3,     'Magnitude of dissolve effect')
config.text('\n')

config.text('  *** Weapon-specific options ***')
config.text('')
config.text('Use the following server command to specify weapon-specific dissolve settings:')
config.text('')
config.text('entdissolver_addspecial <name> <delay> [magnitude] ["dissolvetypes"]')
config.text('')
config.text('See the bottom of these options for a list of acceptable weapon names.')
config.text('\n')

config.text('Place your entdissolver_addspecial commands below:')
config.command('entdissolver_addspecial')
config.text('\n')

config.text('Uncomment the following example options if you wish to use them:')
config.text('')
config.text('entdissolver_addspecial awp 6 20')
config.text('   Uncomment the above line to dissolve AWPs after six seconds with a magitude of 20')
config.text('')
config.text('entdissolver_addspecial m3 18')
config.text('   Uncomment the above line to dissolve M3s after 18 seconds')
config.text('')
config.text('entdissolver_addspecial tmp 15 3 "02"')
config.text('   Uncomment the above line to dissolve TMPs after 15 seconds with a magnitude of three and the dissolve type of zero or two')
config.text('\n')

config.text('Weapon-specific dissolve settings can be removed with the following server command:')
config.text('')
config.text('entdissolver_removespecial <name>')
config.text('')
config.text('Due to the fact weapon-specific dissolve settings can be added or removed removed above, this command will not be used by most users.')
config.text('\n')

config.text('  *** Acceptable weapon names ***')
config.text('These are the weapon names acceptable in the weapon-specific dissolve settings category above along with their more common in-game names.')
config.text('This is just a reference--these are not options.')
config.text('')
config.text('deagle      - Night Hawk .50C')
config.text('ak47        - CV-47')
config.text('scout       - Schmidt Scout')
config.text('aug         - Bullpup')
config.text('g3sg1       - D3/AU1')
config.text('galil       - IDF Defender')
config.text('famas       - Clarion 5.56')
config.text('m4a1        - Maverick M4A1 Carbine')
config.text('sg552       - Krieg 552')
config.text('sg550       - Krieg 550 Commando')
config.text('m249        - M249')
config.text('awp         - Magnum Sniper Rifle')
config.text('tmp         - Schmidt Machine Pistol')
config.text('mp5navy     - KM Sub-Machine Gun')
config.text('glock       - 9x19mm Sidearm')
config.text('elite       - .40 Dual Elites')
config.text('m3          - Leone 12 Gauge Super')
config.text('xm1014      - Leone YG1265 Auto Shotgun')
config.text('usp         - KM .45 Tactical')
config.text('mac10       - Ingram MAC-10')
config.text('ump45       - KM UMP45')
config.text('p228        - 228 Compact')
config.text('fiveseven   - ES Five-Seven')
config.text('p90         - ES C90')
config.text('hegrenade   - HE Grenade')
config.text('knife       - Knife')
config.text('\n')

config.text('***** Command documentation *****')
config.text('')
config.text('This addon provides the following server commands while loaded:')
config.text('')
config.text('entdissolver_dissolve <index> <dissolvetype> <magnitude>')
config.text('- Dissolves an entity by index given a dissolve type and magnitude')
config.text('   index         - Index of entity to dissolve')
config.text('   dissolvetype  - Dissolve effect to use: 0 = Energy, 1 = Heavy electrical, 2 = Light electrical, 3 = Core effect')
config.text('   magnitude     - Desired magnitude of dissolve effect')
config.text('')
config.text('entdissolver_dissolve_delay <delay> <index> <dissolvetype> <magnitude>')
config.text('- Dissolves an entity by index after a delay')
config.text('   delay         - Number of seconds to wait before dissolving entity')
config.text('   index         - Index of entity to dissolve')
config.text('   dissolvetype  - Dissolve effect to use: 0 = Energy, 1 = Heavy electrical, 2 = Light electrical, 3 = Core effect')
config.text('   magnitude     - Desired magnitude of dissolve effect')

config.write()

config.execute()

###

class Dissolver(object):
   def __init__(self):
      self.delays = set()

   def dissolve(self, target_index, dissolvetype, magnitude):
      """ Dissolve an entity by index given a dissolve type and magnitude """
      userid = es.getuserid()
      if not userid:
         return
      # Mark the passed index as the target entity
      es.entitysetvalue(target_index, 'targetname', 'dissolve_me')

      # Create an env_entity_dissolver for dissolving the entity
      index = es.createentity('env_entity_dissolver')
      es.entitysetvalue(index, 'targetname', 'entdissolver_dissolve') # Unique name
      es.entitysetvalue(index, 'target', 'dissolve_me') # Dissolve target
      es.entitysetvalue(index, 'magnitude', magnitude) # Dissolve effect magnitude
      es.entitysetvalue(index, 'dissolvetype', dissolvetype) # Dissolve effect type (0-4)
      es.spawnentity(index)

      # Dissolve the target entity
      es.fire(userid, 'entdissolver_dissolve', 'Dissolve')
      es.fire(userid, 'entdissolver_dissolve', 'Kill')

      # If we were watching for this entity, it's gone now
      ragdolls.remove(target_index)
      weapons.remove(target_index)
      items.remove(target_index)

   def dissolveDelay(self, delay, target_index, dissolvetype, magnitude):
      """ Dissolves an entity after a delay """
      self.delays.add(target_index)
      gamethread.delayedname(delay, 'entdissolver_' + str(target_index), self._dissolveDelayStep,
       (target_index, dissolvetype, magnitude, es.createentitylist(target_index)[target_index]['classname']))

   def _dissolveDelayStep(self, target_index, dissolvetype, magnitude, classname):
      """ Step two for dissolving entities on a delay """
      if target_index in self.delays:
         self.delays.remove(target_index)

      entlist = es.createentitylist(target_index)
      # The entity has to exist, be the same class, and have no owner
      if (entlist and entlist[target_index]['classname'] == classname and
       es.getindexprop(target_index, 'CBaseEntity.m_hOwnerEntity') == -1):
         self.dissolve(target_index, dissolvetype, magnitude)

   def removeDelay(self, index):
      """ Cancel the index delay and remove the index from our delay list """
      if index in self.delays:
         self.delays.remove(index)
      gamethread.cancelDelayed('entdissolver_' + str(index))

   def cancelDelays(self):
      """ Cancel outstanding delays """
      for index in self.delays:
         gamethread.cancelDelayed('entdissolver_' + str(index))
      self.delays.clear()

      # We sometimes use this, too
      gamethread.cancelDelayed('entdissolver')

entd = Dissolver()


def dissolve(*a, **kw):
   entd.dissolve(*a, **kw)
dissolve.__doc__ = entd.dissolve.__doc__

def dissolveDelay(*a, **kw):
   entd.dissolveDelay(*a, **kw)
dissolveDelay.__doc__ = entd.dissolveDelay.__doc__

def removeDelay(*a, **kw):
   entd.removeDelay(*a, **kw)
removeDelay.__doc__ = entd.removeDelay.__doc__

###

def dissolve_cmd(args):
   """
   entdissolver_dissolve <index> <dissolvetype> <magnitude>
   """
   if len(args) == 3:
      dissolve(*args)

   else:
      es.dbgmsg(0, 'Syntax: entdissolver_dissolve <index> <dissolvetype> <magnitude>')

cmdlib.registerServerCommand('entdissolver_dissolve', dissolve_cmd,
 'Dissolves an entity with env_entity_dissolver\nSyntax: entdissolver_dissolve <index> <dissolvetype> <magnitude>')


def dissolve_delay_cmd(args):
   """
   entdissolver_dissolve_delay <delay> <index> <dissolvetype> <magnitude>
   """
   if len(args) == 4:
      dissolveDelay(*args)

   else:
      es.dbgmsg(0, 'Syntax: entdissolver_dissolve_delay <delay> <index> <dissolvetype> <magnitude>')

cmdlib.registerServerCommand('entdissolver_dissolve_delay', dissolve_delay_cmd,
 'Dissolves an entity with env_entity_dissolver after a certain amount of time\nSyntax: entdissolver_dissolve_delay <delay> <index> <dissolvetype> <magnitude>')

###

class RagdollDissolver(object):
   ragdoll_names = None

   def __init__(self):
      self.clear()

   def clear(self):
      self.ragdolls = set(self.getRagdollIndexes())

   def add(self):
      # If we aren't removing ragdolls, we're done
      if not cvar_ragdoll_delay:
         return

      # Look through all ragdolls
      for index, howner in self.getRagdollIndexes():
         if index not in self.ragdolls:
            # Apply player score to magnitude, if necessary
            magnitude = cvar_ragdoll_magnitude
            if cvar_ragdoll_magnitude_score:
               userid = es.getuserid(howner)
               if userid:
                  magnitude = max(1, magnitude * (playerlib.getPlayer(userid).kills + 1))
            dissolveDelay(cvar_ragdoll_delay, index, random.choice(str(cvar_ragdoll_type)), magnitude)
            self.ragdolls.add(index)

   def remove(self, index):
      if index in self.ragdolls:
         self.ragdolls.remove(index)

   def getRagdollIndexes(self):
      try:
         entname, propname = self.getRagdollName()
      except NameError:
         raise StopIteration # No elements

      # Create the entity list
      entlist = es.createentitylist(entname)
      if propname is None: # No owner property found
         for index in entlist:
            yield index, None
      else: # If we found the owner property, return owners
         for index in entlist:
            yield index, entlist[index][propname]
      raise StopIteration

   def getRagdollName(self):
      if not self.ragdoll_names is None:
         return self.ragdoll_names

      entlist = es.createentitylist()
      for index in entlist:
         e = entlist[index]
         classname = e['classname']
         if classname.endswith('_ragdoll'):
            for p in e:
               if p.endswith('m_hPlayer'):
                  x = self.ragdoll_names = classname, p
                  return x
            x = self.ragdoll_names = classname, None
            return x
      raise NameError, 'Unable to find ragdoll entity'

ragdolls = RagdollDissolver()


class WeaponDissolver(object):
   class SpecialDissolve(object):
      def __init__(self, name, delay=cvar_weapon_delay,
       magnitude=cvar_weapon_magnitude, dissolvetypes=cvar_weapon_type):
         self.name  = name
         self.delay = delay
         self.type  = dissolvetypes
         self.magnitude = magnitude

   def __init__(self):
      self.weapons = set()
      self.clear()

      self.special = {}

   def clear(self):
      self.weapons.clear()
      for index in weaponlib.xgetIndexList('#all'):
         if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == -1:
            self.weapons.add(index)

   def add(self, weapon='!objective'):
      if not cvar_weapon_delay:
         return

      for index in weaponlib.xgetIndexList(weapon):
         if index not in self.weapons and es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == -1:
            # Does this weapon have unique settings?
            classname = es.entitygetvalue(index, 'classname')
            if classname in self.special:
               s = self.special[classname]
               dissolveDelay(s.delay, index, random.choice(str(s.type)), s.magnitude)
            # Standard dissolve
            else:
               dissolveDelay(cvar_weapon_delay, index, random.choice(str(cvar_weapon_type)), cvar_weapon_magnitude)
            self.weapons.add(index)

   def remove(self, index):
      if index in self.weapons:
         self.weapons.remove(index)

   def weaponPickup(self, weapon):
      for index in weapon.indexlist:
         if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') != -1:
            if index in entd.delays:
               entd.removeDelay(index)
            self.remove(index)

   def addSpecial(self, name, *a, **kw):
      special[name] = SpecialDissolve(name, *a, **kw)

   def removeSpecial(self, name):
      if name in special:
         del special[name]

weapons = WeaponDissolver()


class ItemDissolver(object):
   def __init__(self):
      self.clear()

   def clear(self):
      self.items = set(self.getEntityList())

   def add(self):
      if not cvar_item_delay or not self.canDissolveDefusers():
         return

      for index in self.getEntityList():
         if index not in self.items:
            dissolveDelay(cvar_item_delay, index, random.choice(str(cvar_item_type)), cvar_item_magnitude)
            self.items.add(index)

   def remove(self, index):
      if index in self.items:
         self.items.remove(index)

   @staticmethod
   def getEntityList():
      return iter(es.createentitylist('item_defuser'))

   @staticmethod
   def canDissolveDefusers():
      return all(player.hasDefuser() for player in playerlib.getPlayerList('#ct,#alive'))

items = ItemDissolver()


def cleanupAll():
   """ Marks new idle ragdolls, weapons and items to be dissolved """
   ragdolls.add()
   weapons.add()
   items.add()

###

def addspecial_cmd(args):
   """
   entdissolver_addspecial <name> <delay> [magnitude] ["dissolvetypes"]
   """
   if len(args) in (2, 3, 4):
      weapon = weaponlib.getWeapon(args[0])
      if weapon is None:
         es.dbgmsg(0, 'EntDissolver: Unknown weapon type "' + args[0] + '"')
         return

      items.addSpecial(str(weapon), *args[1:])

   else:
      es.dbgmsg(0, 'Syntax: entdissolver_addspecial <name> <delay> [magnitude] ["dissolvetypes"]')

cmdlib.registerServerCommand('entdissolver_addspecial', addspecial_cmd,
 'Adds unique dissolve settings for a specific weapon\nSyntax: entdissolver_addspecial <name> <delay> [magnitude] ["dissolvetypes"]')


def removespecial_cmd(args):
   """
   entdissolver_removespecial <name>
   """
   if len(args) == 1:
      weapon = weaponlib.getWeapon(args[0])
      if weapon is None:
         es.dbgmsg(0, 'EntDissolver: Unknown weapon type "' + args[0] + '"')
         return

      items.removeSpecial(str(weapon))

   else:
      es.dbgmsg(0, 'Syntax: entdissolver_removespecial <name>')

cmdlib.registerServerCommand('entdissolver_removespecial', removespecial_cmd,
 'Removes unique dissolve settings for a specific weapon\nSyntax: entdissolver_removespecial <name>')

###

def ccFilter(userid, args):
   if args[0].lower() == 'drop':
      gamethread.delayedname(0, 'entdissolver', weapons.add, playerlib.getPlayer(userid).weapon)
   return True

es.addons.registerClientCommandFilter(ccFilter)

###

def round_start(event_var):
   ragdolls.clear()
   weapons.clear()
   items.clear()
   entd.cancelDelays()


def player_death(event_var):
   gamethread.delayedname(0, 'entdissolver', cleanupAll)


def item_pickup(event_var):
   weapon = weaponlib.getWeapon(event_var['item'])
   # If we recognize this weapon we need to remove any delays it has
   if not weapon is None:
      weapons.weaponPickup(weapon)


def unload():
   entd.cancelDelays()

   cmdlib.unregisterServerCommand('entdissolver_dissolve')
   cmdlib.unregisterServerCommand('entdissolver_dissolve_delay')

   es.addons.unregisterClientCommandFilter(ccFilter)

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)